"""
Excel (XLSX) Input filter for ingest_json.

Processes Excel XLSX files with [openpyxl](https://openpyxl.readthedocs.io/en/stable/).

Configuration values:

* `worksheet` (default: `None`): The name of the worksheet to process. If `None`, the default one will be used.
* `rows` (default: `dict`): type of elements in the result list:
    * `dict`: elements will be dictionaries, with the keys taken from the `header-row`.
    * `list`: each resulting row will be an array values.
* `header-row` (default: `0`): if `rows` is `dict`, the (0-based) index of the header row. All rows before the
    header row will be skipped.
* `skip-rows` (default: `0`): number of rows to skip at the beginning of the file (apart from the header and pre-header
    ones if `rows` is `dict`).
* `skip-empty-rows` (default: `True`): whether to omit empty rows (i.e., those with no values) from the result.
"""
from __future__ import annotations

from datetime import datetime
from io import BytesIO
from typing import Any
from openpyxl import load_workbook
from openpyxl.cell import Cell

from ogc.na import util

DEFAULT_CONF = {
    'worksheet': None,
    'rows': 'dict',
    'header-row': 0,
    'skip-rows': 0,
    'skip-empty-rows': True,
}


def _cell_to_json(c: Cell) -> str | float | int | None:
    if isinstance(c.value, datetime):
        return c.value.isoformat()
    return c.value


def apply_filter(content: bytes, conf: dict[str, Any] | None) -> tuple[dict[str, Any] | list, dict[str, Any] | None]:
    conf = util.deep_update(DEFAULT_CONF, conf) if conf else DEFAULT_CONF

    metadata = {
        'filter': {
            'conf': conf,
        },
    }

    wb = load_workbook(filename=BytesIO(content), read_only=True)
    if conf['worksheet']:
        ws = wb[conf['worksheet']]
    else:
        ws = wb.worksheets[0]
    rows = ws.rows
    metadata['worksheet'] = ws.title

    headers = None
    if conf['rows'] == 'dict':
        header_row = max(conf['header-row'], 0)
        # Skip to header row
        for i in range(header_row):
            next(rows, None)
        headers = next(rows, [])
        if not headers:
            return [], None
        else:
            headers = [_cell_to_json(c) for c in headers]
        metadata['headers'] = headers

    for i in range(conf['skip-rows']):
        next(rows, None)

    result = []
    for row in rows:
        row_values = [_cell_to_json(c) for c in row]
        if conf['skip-empty-rows'] and not any(c is not None for c in row_values):
            # skip empty rows
            continue
        if conf['rows'] == 'list':
            result.append(row_values)
        else:
            result.append(dict(zip(headers, row_values)))

    return result, metadata



"""
SPARQL Graph Store Protocol operations
"""
from __future__ import annotations

import argparse
import getpass
import sys
from io import IOBase
from pathlib import Path
from typing import Generator, IO

import requests
from rdflib import Graph


class GraphStore:
    """
    Encapsulates Graph Store Protocol configuration for executing operations.
    """

    def __init__(self, url: str, auth_details: tuple[str, str] | None = None):
        """
        Constructs a new GraphStore
        :param url: SPARQL Graph Store Protocol URL
        :param auth_details: tuple in the form ('username', 'password') for authentication
        """
        self._url = url
        self.auth_details = auth_details
        self.put = self.replace
        self.post = self.add

    def _post_or_put(self, method: str, graph_uri, source: str | bytes | Path, format: str = 'text/turtle'):
        if graph_uri:
            params = {'graph': str(graph_uri)}
        else:
            params = 'default'

        if isinstance(source, Path):
            with open(source, 'r') as f:
                data = f.read()
        elif isinstance(source, IO) or isinstance(source, IOBase):
            data = source.read()
        else:
            data = source

        if isinstance(data, str):
            data = data.encode('utf-8')

        method = getattr(requests, method)
        r = method(
            self._url,
            params=params,
            headers={
                'Content-type': format,
            },
            auth=self.auth_details,
            data=data,
        )

        if isinstance(data, Generator):
            data.close()

        r.raise_for_status()

    def delete(self, graph_uri: str | None, ignore_404=True):
        """
        Deletes a graph from the Graph Store
        :param graph_uri: URI for the graph (if None, the default graph will be used)
        :param ignore_404: Whether to ignore HTTP 404 errors (otherwise, an Exception will be thrown)
        :return:
        """
        if graph_uri:
            params = {'graph': str(graph_uri)}
        else:
            params = 'default'

        r = requests.delete(
            self._url,
            params=params,
            auth=self.auth_details,
        )
        if not (ignore_404 and r.status_code == 404):
            r.raise_for_status()

    def replace(self, graph_uri: str | None, source: str | bytes | Path, format: str = 'text/turtle'):
        """
        Replaces the data for a Graph Store graph with the provided source (HTTP PUT operation)
        :param graph_uri: URI for the graph (if None, the default graph will be used)
        :param source: Source data or file name
        :param format: Media type to provide to the Graph Store
        :return:
        """
        self._post_or_put('put', graph_uri, source, format)

    def add(self, graph_uri: str | None, source: str | bytes | Path, format: str = 'text/turtle'):
        """
        Adds data from the provided source to a Graph Store graph (HTTP PUT operation)
        :param graph_uri: URI for the graph (if None, the default graph will be used)
        :param source: Source data or file name
        :param format: Media type to provide to the Graph Store
        :return:
        """
        self._post_or_put('post', graph_uri, source, format)

    def get(self, graph_uri) -> Graph:
        """
        Retrieves the data from a graph in the Graph Store
        :param graph_uri: URI for the graph (if None, the default graph will be used)
        :return: An [RDFLib Graph][rdflib.Graph]
        """
        if graph_uri:
            params = {'graph': str(graph_uri)}
        else:
            params = 'default'

        r = requests.get(
            self._url,
            params=params,
            auth=self.auth_details,
        )
        r.raise_for_status()
        g = Graph().parse(r.content)
        return g


def _main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        'graph_store',
        metavar='GRAPH-STORE',
        help='Graph Store Protocol endpoint',
    )

    parser.add_argument(
        'operation',
        choices=['get', 'add', 'replace', 'delete', 'post', 'put'],
        help='Operation to perform'
    )

    parser.add_argument(
        'graph_uri',
        metavar='GRAPH-URI',
        nargs='?',
        help='Graph URI',
    )

    parser.add_argument(
        '--data',
        help='Data for add/post and replace/put operations. If "-", read from stdin',
    )

    parser.add_argument(
        '--data-file',
        help='Data file for add/post and replace/put operations. If "-", read from stdin',
    )

    parser.add_argument(
        '-u',
        '--username',
        help='Username for HTTP authentication',
    )

    parser.add_argument(
        '-p',
        '--password',
        help='Password for HTTP authentication',
    )

    parser.add_argument(
        '-P',
        '--request-password',
        action='store_true',
        help='Request password from the user interactively',
    )

    args = parser.parse_args()

    if args.request_password:
        password = getpass.getpass("Password: ")
    else:
        password = args.password

    auth = (args.username, password) if args.username else None
    gs = GraphStore(args.graph_store, auth_details=auth)
    if args.operation in ('add', 'replace', 'post', 'put'):
        if not (args.data or args.data_file):
            parser.error('Operation requires either --data or --data-file')
        if args.data and args.data_file:
            parser.error('Only one of --data, --data-file must be provided')
        if args.data == '-':
            data = sys.stdin
        elif args.data:
            data = args.data
        else:
            data = Path(args.data_file)

        getattr(gs, args.operation)(args.graph_uri, source=data)
    elif args.operation == 'delete':
        gs.delete(args.graph_uri)
    else:
        print(gs.get(args.graph_uri).serialize(format='ttl'))


if __name__ == '__main__':
    _main()



"""
XML Input filter for ingest_json.

Processes XML files with [xmltodict](https://pypi.org/project/xmltodict/). Attributes are prefixed with `_` instead
of `@` by default.

Configuration values:

* `process-namespaces` (default: `False`): Whether to process and expand namespaces (see xmltodict documentation)
* `namespaces` (default: `None`): Namespace to prefix mappings dict in `url: prefix` format.
* `attr-prefix` (default: `_`): Prefix that will be used for attributes (in order to avoid potential clashes with
  element names).
* `namespace-separator` (default `:`): String that will be used to separate the namespace prefix and the local name
  when processing namespaces.
* `text-property` (default: `_`): property name that will be used to put the element's text content into
* `disable-entities` (default: `True`): disable processing of XML entities (to avoid XXE injections when parsing
  untrusted data)
"""
from __future__ import annotations

from io import StringIO
from typing import Any
import xmltodict

from ogc.na import util

DEFAULT_CONF = {
    'process-namespaces': False,
    'namespaces': None,
    'attr-prefix': '_',
    'namespace-separator': ':',
    'text-property': '_',
    'disable-entities': True,
}


def apply_filter(content: bytes, conf: dict[str, Any] | None) -> tuple[dict[str, Any] | list, dict[str, Any] | None]:
    conf = util.deep_update(DEFAULT_CONF, conf) if conf else DEFAULT_CONF

    metadata = {
        'filter': {
            'conf': conf,
        },
    }

    textio = StringIO(content.decode('utf-8'))
    result = xmltodict.parse(textio.read(),
                             process_namespaces=conf['process-namespaces'],
                             namespaces=conf['namespaces'],
                             attr_prefix=conf['attr-prefix'],
                             namespace_separator=conf['namespace-separator'],
                             cdata_key=conf['text-property'],
                             disable_entities=conf['disable-entities'],
                             )

    return result, metadata
